"""
Create a new django migration with support for fixing conflicts.
"""

import os
import pathlib
from functools import partial

from django.apps import apps
from django.conf import settings
from django.core.management.base import CommandError
from django.core.management.commands.makemigrations import Command as BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections, router
from django.db.migrations.loader import MigrationLoader
from git import InvalidGitRepositoryError, Repo

from migration_fixer.utils import (
    fix_numbered_migration,
    get_filename,
    migration_sorter,
    no_translations,
)


class Command(BaseCommand):
    """
    Create a new django migration with support for fixing conflicts.
    """

    success_msg = "Successfully fixed migrations."

    def __init__(self, *args, repo=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.cwd = os.getcwd()
        self.repo = repo or Repo.init(self.cwd)

    def add_arguments(self, parser):
        parser.add_argument(
            "--fix",
            action="store_true",
            help="Fix migrations conflicts.",
        )
        parser.add_argument(
            "-b",
            "--default-branch",
            help="The name of the default branch.",
            default="main",
        )
        parser.add_argument(
            "-r",
            "--remote",
            help="Git remote.",
            default="origin",
        )
        parser.add_argument(
            "-f",
            "--force-update",
            help="Force update the default branch.",
            action="store_true",
        )
        super().add_arguments(parser)

    @no_translations
    def handle(self, *app_labels, **options):
        self.merge = options["merge"]
        self.fix = options["fix"]
        self.force_update = options["force_update"]
        self.default_branch = options["default_branch"]
        self.remote = options["remote"]

        if self.fix:
            try:
                super().handle(*app_labels, **options)
            except CommandError as e:
                [message] = e.args
                if "Conflicting migrations" in message:
                    if self.verbosity >= 2:
                        self.stdout.write("Verifying git repository...")

                    try:
                        self.repo.git_dir and self.repo.head.commit
                    except (ValueError, InvalidGitRepositoryError):
                        is_git_repo = False
                    else:
                        is_git_repo = True

                    if not is_git_repo:
                        raise CommandError(
                            self.style.ERROR(
                                f"Git repository is not yet setup. "
                                "Please run (git init) in"
                                f'\n"{self.cwd}"'
                            )
                        )

                    if self.verbosity >= 2:
                        self.stdout.write("Retrieving the current branch...")

                    current_branch = self.repo.head.ref.name

                    if self.repo.is_dirty():  # pragma: no cover
                        raise CommandError(
                            self.style.ERROR(
                                "Git repository has uncommitted changes. "
                                "Please commit any outstanding changes."
                            )
                        )

                    if self.verbosity >= 2:
                        self.stdout.write(
                            f"Fetching git remote origin changes on: {self.default_branch}"
                        )

                    if current_branch == self.default_branch:  # pragma: no cover
                        remote = self.repo.remotes[self.remote]
                        remote.pull(
                            self.default_branch,
                            force=self.force_update,
                        )
                    else:
                        remote = self.repo.remotes[self.remote]
                        remote.fetch(
                            f"{self.default_branch}:{self.default_branch}",
                            force=self.force_update,
                        )

                    if self.verbosity >= 2:
                        self.stdout.write(
                            f"Retrieving the last commit sha on: {self.default_branch}"
                        )

                    default_branch_commit = self.repo.commit(self.default_branch)

                    current_commit = self.repo.commit(current_branch)

                    # Load the current graph state. Pass in None for the connection so
                    # the loader doesn't try to resolve replaced migrations from DB.
                    loader = MigrationLoader(None, ignore_no_migrations=True)

                    # Raise an error if any migrations are applied before their dependencies.
                    consistency_check_labels = {
                        config.label for config in apps.get_app_configs()
                    }
                    # Non-default databases are only checked if database routers used.
                    aliases_to_check = (
                        connections if settings.DATABASE_ROUTERS else [DEFAULT_DB_ALIAS]
                    )
                    for alias in sorted(aliases_to_check):
                        connection = connections[alias]
                        if connection.settings_dict[
                            "ENGINE"
                        ] != "django.db.backends.dummy" and any(
                            # At least one model must be migrated to the database.
                            router.allow_migrate(
                                connection.alias,
                                app_label,
                                model_name=model._meta.object_name,
                            )
                            for app_label in consistency_check_labels
                            for model in apps.get_app_config(app_label).get_models()
                        ):
                            loader.check_consistent_history(connection)

                    # Before anything else, see if there's conflicting apps and drop out
                    # hard if there are any and they don't want to merge
                    conflicts = loader.detect_conflicts()

                    for app_label in conflicts:
                        conflict = conflicts.get(app_label)
                        migration_module, _ = loader.migrations_module(app_label)
                        migration_absolute_path = os.path.join(
                            *migration_module.split(".")
                        )
                        migration_path = pathlib.Path(
                            os.path.join(settings.BASE_DIR, migration_absolute_path)
                        )

                        with migration_path:
                            if self.verbosity >= 2:
                                self.stdout.write(
                                    "Retrieving changed files between "
                                    f"the current branch and {self.default_branch}"
                                )

                            try:
                                diff_index = default_branch_commit.diff(current_commit)

                                # Files different on the current branch
                                changed_files = [
                                    diff.b_path
                                    for diff in diff_index
                                    if (
                                        migration_absolute_path
                                        in getattr(diff.a_blob, "abspath", "")
                                        or migration_absolute_path
                                        in getattr(diff.b_blob, "abspath", "")
                                    )
                                ]

                                # Only consider files from the current conflict.
                                conflict_base = [
                                    get_filename(path)
                                    for path in changed_files
                                    if get_filename(path) in conflict
                                ][0]

                                sorted_changed_files = sorted(
                                    changed_files,
                                    key=partial(migration_sorter, app_label=app_label),
                                )

                                changed_files = [
                                    path
                                    for path in sorted_changed_files
                                    if (
                                        int(conflict_base.split("_")[0])
                                        >= int(get_filename(path).split("_")[0])
                                    )
                                ]

                                # Local migration
                                local_filenames = [
                                    get_filename(p) for p in changed_files
                                ]
                                if self.verbosity >= 2:
                                    self.stdout.write(
                                        f"Retrieving the last migration on: {self.default_branch}"
                                    )

                                last_remote = [
                                    name
                                    for name in conflict
                                    if name not in local_filenames
                                ]

                                if not last_remote:  # pragma: no cover
                                    raise CommandError(
                                        self.style.ERROR(
                                            f"Unable to determine the last migration on: "
                                            f"{self.default_branch}. "
                                            "Please verify the target branch using"
                                            '"-b [target branch]".',
                                        )
                                    )

                                last_remote_filename, *rest = last_remote
                                changed_files = changed_files or [
                                    f"{fname}.py" for fname in rest
                                ]

                                seed_split = last_remote_filename.split("_")

                                if (
                                    seed_split
                                    and len(seed_split) > 1
                                    and str(seed_split[0]).isdigit()
                                ):
                                    if self.verbosity >= 2:
                                        self.stdout.write(
                                            "Fixing numbered migration..."
                                        )

                                    fix_numbered_migration(
                                        app_label=app_label,
                                        migration_path=migration_path,
                                        seed=int(seed_split[0]),
                                        start_name=last_remote_filename,
                                        changed_files=changed_files,
                                        writer=(
                                            lambda message: self.stdout.write(message)
                                            if self.verbosity >= 2
                                            else lambda x: x
                                        ),
                                    )
                                else:  # pragma: no cover
                                    raise ValueError(
                                        f"Unable to fix migration: {last_remote_filename}. \n"
                                        f"NOTE: It needs to begin with a number. eg. 0001_*",
                                    )
                            except (ValueError, IndexError, TypeError) as e:
                                self.stderr.write(f"Error: {e}")
                            else:
                                self.stdout.write(self.success_msg)

        else:
            return super(Command, self).handle(*app_labels, **options)



DEFAULT_BRANCH = "main"
# Single merge to main
TEST_01_MIGRATION_BRANCH = "feature/migration-test-01"
# Single merge to main after 01 is merged (single conflict)
TEST_02_MIGRATION_BRANCH = "feature/migration-test-02"
# Single merge to main after 01 is merged (multiple conflicts)
TEST_03_MIGRATION_BRANCH = "feature/migration-test-03"
# Single merge to main after 01 is merged (named module)
TEST_04_MIGRATION_BRANCH = "feature/migration-test-04"
# Multiple migrations with a named migration file
TEST_05_MIGRATION_BRANCH = "feature/migration-test-05"



import os
import re
from itertools import count
from pathlib import Path
from typing import Callable, List

DEFAULT_TIMEOUT = 120
MIGRATION_REGEX = "\\((?P<comma>['\"]){app_label}(['\"]),\\s(['\"])(?P<conflict_migration>.*)(['\"])\\),"


def _clean_message(output: str) -> str:
    """Strips the succeeding new line and carriage return characters."""
    return output.rstrip("\n\r")


def _decode_message(output: bytes, encoding: str) -> str:
    """Converts bytes to string, stripping white spaces."""
    return output.decode(encoding).strip()


def _update_migration(conflict_path: Path, app_label: str, prev_migration: str) -> None:
    """Modify the migration file."""
    regex = MIGRATION_REGEX.format(app_label=app_label)
    replace_regex = re.compile(regex, re.I)

    match = replace_regex.search(conflict_path.read_text())

    if match:
        comma = match.group("comma")
        replacement = f"({comma}{app_label}{comma}, {comma}{prev_migration}{comma}),"

        # Update the migration
        output = re.sub(
            replace_regex,
            replacement,
            conflict_path.read_text(),
        )

        # Write to the conflict file.
        conflict_path.write_text(output)
    else:  # pragma: no cover
        raise ValueError(f'Couldn\'t find "{regex}" in {conflict_path.name}')


def migration_sorter(path: str, app_label: str) -> int:
    path = os.path.split(path)[-1]
    parts = path.split("_")
    key = parts[0]

    if not str(key).isdigit():
        raise ValueError(
            f'Unable to fix migration for "{app_label}" app: {path}\n'
            f"NOTE: It needs to begin with a number. eg. 0001_*",
        )

    return int(key)


def fix_numbered_migration(
    *,
    app_label: str,
    migration_path: Path,
    seed: int,
    start_name: str,
    changed_files: List[str],
    writer: Callable[[str], None],
) -> None:
    """Resolve migration conflicts for numbered migrations."""
    seen = [start_name]
    counter = count(seed + 1)  # 0537 -> 538

    for path in changed_files:
        next_ = str(next(counter))

        if len(next_) < 4:
            next_ = f"{next_}".rjust(4, "0")  # 0537

        basename = os.path.basename(path)
        conflict_path = migration_path / basename
        conflict_parts = basename.split("_")

        conflict_parts[0] = next_

        new_conflict_name = "_".join(conflict_parts)

        conflict_new_path = conflict_path.with_name(new_conflict_name)

        with conflict_path:
            prev_migration = seen[-1]

            writer(
                f'Updating migration "{conflict_path.name}" dependency to {prev_migration}'
            )
            _update_migration(conflict_path, app_label, prev_migration)

            writer(
                f'Renaming migration "{conflict_path.name}" to "{conflict_new_path.name}"'
            )
            # Rename the migration file
            conflict_path.rename(conflict_new_path)

            seen.append(conflict_new_path.stem)


def no_translations(handle_func):
    """Decorator that forces a command to run with translations deactivated."""

    def wrapped(*args, **kwargs):
        from django.utils import translation

        saved_locale = translation.get_language()
        translation.deactivate_all()
        try:
            res = handle_func(*args, **kwargs)
        finally:
            if saved_locale is not None:
                translation.activate(saved_locale)
        return res

    return wrapped


def get_filename(path: str) -> str:
    """Return the file name from a path."""
    return os.path.splitext(os.path.basename(path))[0]



"""Tests for `migration_fixer` package."""

import pytest
from django.core.management.base import CommandError

from migration_fixer.management.commands.makemigrations import Command
from migration_fixer.tests.management.commands._constants import (
    TEST_01_MIGRATION_BRANCH,
    TEST_02_MIGRATION_BRANCH,
    TEST_03_MIGRATION_BRANCH,
    TEST_04_MIGRATION_BRANCH,
    TEST_05_MIGRATION_BRANCH,
)
from migration_fixer.tests.management.commands._utils import (
    execute_command,
    temporary_checkout,
)


@pytest.mark.env("invalid_repo")
@pytest.mark.django_db
def test_invalid_repo(invalid_git_repo, mocker):
    mocker.patch(
        "django.core.management.commands.makemigrations.Command.handle",
        side_effect=CommandError("Conflicting migrations detected"),
    )
    cmd = Command(repo=invalid_git_repo.api)
    cmd.verbosity = 1

    with pytest.raises(CommandError) as exc_info:
        execute_command(cmd, fix=True)

    assert "Git repository is not yet setup." in str(exc_info.value)


@pytest.mark.env("test_01")
@pytest.mark.django_db
def test_run_makemigrations_is_valid_without_any_conflicts(git_repo):
    cmd = Command(repo=git_repo.api)

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_01_MIGRATION_BRANCH,
    ) as target_branch:
        output1 = execute_command(cmd)
        output2 = execute_command(cmd, fix=True)

    assert target_branch.name == TEST_01_MIGRATION_BRANCH
    assert output1 == "No changes detected\n"
    assert output2 == "No changes detected\n"


@pytest.mark.env("test_01")
@pytest.mark.django_db
def test_run_makemigrations_is_valid_without_any_conflicts_verbose(git_repo):
    cmd = Command(repo=git_repo.api)

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_01_MIGRATION_BRANCH,
    ) as target_branch:
        output1 = execute_command(cmd, verbosity=2)
        output2 = execute_command(cmd, verbosity=2, fix=True)

    assert target_branch.name == TEST_01_MIGRATION_BRANCH
    assert output1 == "No changes detected\n"
    assert output2 == "No changes detected\n"


@pytest.mark.env("test_02")
@pytest.mark.django_db
def test_run_makemigrations_with_fix_is_valid_for_conflicts(git_repo):
    cmd = Command(repo=git_repo.api)

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_02_MIGRATION_BRANCH,
    ) as target_branch:
        output = execute_command(cmd, default_branch=TEST_01_MIGRATION_BRANCH, fix=True)

    assert target_branch.name == TEST_02_MIGRATION_BRANCH
    assert output == f"{cmd.success_msg}\n"


@pytest.mark.env("test_02")
@pytest.mark.django_db
def test_run_makemigrations_with_fix_is_valid_for_conflicts_verbose(git_repo):
    cmd = Command(repo=git_repo.api)
    expected_output = f"""Verifying git repository...
Retrieving the current branch...
Fetching git remote origin changes on: {TEST_01_MIGRATION_BRANCH}
Retrieving the last commit sha on: {TEST_01_MIGRATION_BRANCH}
Retrieving changed files between the current branch and {TEST_01_MIGRATION_BRANCH}
Retrieving the last migration on: {TEST_01_MIGRATION_BRANCH}
Fixing numbered migration...
Updating migration "0002_alter_testmodel_active.py" dependency to 0002_alter_testmodel_age
Renaming migration "0002_alter_testmodel_active.py" to "0003_alter_testmodel_active.py"
Successfully fixed migrations.
"""

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_02_MIGRATION_BRANCH,
    ) as target_branch:
        output = execute_command(
            cmd, default_branch=TEST_01_MIGRATION_BRANCH, verbosity=2, fix=True
        )

    assert target_branch.name == TEST_02_MIGRATION_BRANCH
    assert output == expected_output


@pytest.mark.env("test_03")
@pytest.mark.django_db
def test_run_makemigrations_with_fix_is_valid_for_multiple_file_conflicts(git_repo):
    cmd = Command(repo=git_repo.api)

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_03_MIGRATION_BRANCH,
    ) as target_branch:
        output = execute_command(cmd, default_branch=TEST_01_MIGRATION_BRANCH, fix=True)

    assert target_branch.name == TEST_03_MIGRATION_BRANCH
    assert output == f"{cmd.success_msg}\n"


@pytest.mark.env("test_03")
@pytest.mark.django_db
def test_run_makemigrations_with_fix_is_valid_for_multiple_file_conflicts_verbose(
    git_repo,
):
    cmd = Command(repo=git_repo.api)
    expected_output = f"""Verifying git repository...
Retrieving the current branch...
Fetching git remote origin changes on: {TEST_01_MIGRATION_BRANCH}
Retrieving the last commit sha on: {TEST_01_MIGRATION_BRANCH}
Retrieving changed files between the current branch and {TEST_01_MIGRATION_BRANCH}
Retrieving the last migration on: {TEST_01_MIGRATION_BRANCH}
Fixing numbered migration...
Updating migration "0002_testmodel_created_by.py" dependency to 0002_alter_testmodel_age
Renaming migration "0002_testmodel_created_by.py" to "0003_testmodel_created_by.py"
Updating migration "0003_auto_20210708_1317.py" dependency to 0003_testmodel_created_by
Renaming migration "0003_auto_20210708_1317.py" to "0004_auto_20210708_1317.py"
Successfully fixed migrations.
"""

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_03_MIGRATION_BRANCH,
    ) as target_branch:
        output = execute_command(
            cmd, default_branch=TEST_01_MIGRATION_BRANCH, verbosity=2, fix=True
        )

    assert target_branch.name == TEST_03_MIGRATION_BRANCH
    assert output == expected_output


@pytest.mark.env("test_04")
@pytest.mark.django_db
def test_run_makemigrations_fix_with_an_invalid_module(git_repo):
    expected_output = """Error: Unable to fix migration for "demo" app: testmodel_dob.py
NOTE: It needs to begin with a number. eg. 0001_*
"""
    cmd = Command(repo=git_repo.api)

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_04_MIGRATION_BRANCH,
    ) as target_branch:
        output = execute_command(cmd, default_branch=TEST_01_MIGRATION_BRANCH, fix=True)

    assert target_branch.name == TEST_04_MIGRATION_BRANCH
    assert expected_output == output


@pytest.mark.env("test_05")
@pytest.mark.django_db
def test_run_makemigrations_fix_with_mixed_invalid_modules(git_repo):
    expected_output = """Error: Unable to fix migration for "demo" app: custom_migration.py
NOTE: It needs to begin with a number. eg. 0001_*
"""
    cmd = Command(repo=git_repo.api)

    with temporary_checkout(
        git_repo,
        default_branch_name=TEST_01_MIGRATION_BRANCH,
        target_branch_name=TEST_05_MIGRATION_BRANCH,
    ) as target_branch:
        output = execute_command(cmd, default_branch=TEST_01_MIGRATION_BRANCH, fix=True)

    assert target_branch.name == TEST_05_MIGRATION_BRANCH
    assert expected_output == output
